from dash import dcc, html
from dash.dependencies import Input, Output

from ts_app.dash_app import app
from ts_app.dashboard_components import (
    glossary,
    home_page,
    modelling,
    sample,
    upload,
)

server = app.server

app.layout = html.Div(
    [dcc.Location(id="url", refresh=False), html.Div(id="page-content")]
)


@app.callback(Output("page-content", "children"), [Input("url", "pathname")])
def render_page(pathname: str) -> html.Div:
    """Display the page at the given `pathname`.

    Args:
        pathname (str): The url to a page.

    Returns:
        dash.html.Div.Div: Page content.
    """
    if pathname == "/glossary":
        return glossary.layout
    elif pathname == "/sample":
        return generate_dashboard(data_source=sample.input_layout)
    elif pathname == "/upload":
        return generate_dashboard(data_source=upload.input_layout)
    else:
        return home_page.layout


def generate_dashboard(data_source: html.Div) -> html.Div:
    """Get a dashboard layout with the given `data_source`.

    Args:
        data_source (dash.html.Div.Div): A html form to collect input.

    Returns:
        dash.html.Div.Div: Dashboard layout.
    """
    return html.Div(
        id="dashboard-content",
        className="dashboard-content",
        children=[
            # Sidebar with input forms
            html.Div(
                className="side-bar",
                children=[
                    html.Div(data_source, id="data-source"),
                    dcc.Store(id="sample-data-store"),
                    dcc.Store(id="file-upload-store"),
                    modelling.model_param_input,
                ],
            ),
            # Time series forecasting plots and guide
            modelling.graphs_and_guide,
        ],
    )



import plotly.graph_objects as go
from pandas.core.series import Series
from plotly.subplots import make_subplots


def plot_ts_components(
    trend: Series, seasonal: Series, residuals: Series, file_name: str
) -> go.Figure:
    """Get subplots of time series components (trend, seasonal, residuals).

    Args:
        trend (pandas.Series): Estimated trend component.
        seasonal (pandas.Series): Estimated seasonal component.
        residuals (pandas.Series): Estimated residual component.
        file_name (str): Data source information.

    Returns:
        plotly.graph_objs._figure.Figure: Subplots of time series components.
    """
    fig = make_subplots(
        rows=3,
        cols=1,
        shared_xaxes=True,
        subplot_titles=("Trend", "Seasonal", "Residuals"),
        vertical_spacing=0.24,
    )
    fig.add_scatter(
        x=trend.index,
        y=trend,
        mode="lines",
        name="trend",
        line_color="navy",
        row=1,
        col=1,
    )
    fig.add_scatter(
        x=seasonal.index,
        y=seasonal,
        mode="lines",
        name="seasonal",
        line_color="seagreen",
        row=2,
        col=1,
    )
    fig.add_scatter(
        x=residuals.index,
        y=residuals,
        mode="lines",
        name="residuals",
        line_color="#ff3322",
        row=3,
        col=1,
    )
    fig.update_annotations(font_size=12)  # subplot titles are annotations
    fig.update_layout(
        font_family="serif",
        margin={"l": 10, "t": 80, "r": 10, "b": 10},
        paper_bgcolor="#eee",
        plot_bgcolor="#eee",
        showlegend=False,
        title=f"Seasonal Decomposition for {file_name}",
        title_font_size=14,
        xaxis1_showticklabels=True,
        xaxis2_showticklabels=True,
    )
    fig.update_traces(hovertemplate="%{x}: <b>%{y:,.4f}</b>", line_width=1)
    fig.update_xaxes(rangeslider=dict(visible=True, thickness=0.035))
    fig.update_yaxes(fixedrange=True)
    return fig


def plot_forecast(
    actual_data: Series,
    predictions: Series,
    forecast: Series,
    model_info: str,
    file_name: str,
) -> go.Figure:
    """Get a line-plot of the data, along with predicted values and a
    14-period forecast.

    Args:
        actual_data (pandas.Series): The original/input data.
        predictions (pandas.Series): In-sample predicted values.
        forecast (pandas.Series): Out-of-sample forecast values.
        model_info (str): A description of the model's type and order.
        file_name (str): Input data source information.

    Returns:
        plotly.graph_objs._figure.Figure: A line-plot of time series
        forecasting results.
    """
    fig = go.Figure(
        go.Scatter(
            y=actual_data,
            x=actual_data.index,
            mode="lines",
            name="actual data",
            line_color="#8888ff",
        ),
    )
    fig.add_scatter(
        x=predictions.index,
        y=predictions,
        mode="lines",
        name="predictions",
        line_color="navy",
    )
    fig.add_scatter(
        x=forecast.index,
        y=forecast,
        mode="lines",
        name="forecast",
        line_color="lime",
    )

    fig.update_layout(
        font_family="serif",
        hovermode="x unified",
        margin={"l": 10, "t": 80, "r": 10, "b": 10},
        paper_bgcolor="#eee",
        plot_bgcolor="#eee",
        title=f"An {model_info} model fitted on {file_name}",
        title_font_size=13,
    )
    fig.update_traces(hovertemplate="<b>%{y:,.4f}</b>", line_width=1)
    fig.update_xaxes(rangeslider=dict(visible=True, thickness=0.15))
    fig.update_yaxes(fixedrange=True)
    return fig



from dash import dcc, html
from dash.dependencies import Input, Output
from ts_app.dash_app import app
from ts_app.ts_functions import create_arma_sample

input_layout = html.Div(
    id="sample-params",
    className="param-input",
    children=[
        html.H3("Sample parameters"),
        # AR order dropdown
        html.Label("AR Order", htmlFor="sample-ar"),
        dcc.Dropdown(
            id="sample-ar",
            clearable=False,
            placeholder="AR order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(1, 5)],
        ),
        # MA order dropdown
        html.Label("MA Order", htmlFor="sample-ma"),
        dcc.Dropdown(
            id="sample-ma",
            clearable=False,
            placeholder="MA order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(1, 5)],
        ),
    ],
)


@app.callback(
    Output("sample-data-store", "data"),
    [Input("sample-ar", "value"), Input("sample-ma", "value")],
)
def get_sample(ar_order: int, ma_order: int) -> dict:
    """Generate a random ARMA sample with the provided parameters.

    Args:
        ar_order (int): AR order.
        ma_order (int): MA order.

    Returns:
        dict: The sample's description and data.
    """
    sample = create_arma_sample(ar_order, ma_order, size=365)

    return {
        "filename": f"an ARMA({ar_order}, {ma_order}) sample",
        "data": sample.to_json(),
    }



from pkgutil import get_data
from typing import Optional, Tuple

import pandas as pd
import plotly.graph_objs as go
from dash import dcc, html
from dash.dependencies import Input, Output
from statsmodels.tsa.api import seasonal_decompose
from ts_app import plotting
from ts_app.dash_app import app
from ts_app.ts_functions import create_arma_sample, fit_arima_model

summary = get_data("ts_app", "assets/summary.md").decode()

model_param_input = html.Div(
    id="model-params",
    className="param-input",
    children=[
        html.H3("Model parameters"),
        # AR order dropdown
        html.Label("AR Order", htmlFor="model-ar"),
        dcc.Dropdown(
            id="model-ar",
            clearable=False,
            placeholder="AR order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
        # Differencing order dropdown
        html.Label("Differencing Order", htmlFor="model-diff"),
        dcc.Dropdown(
            id="model-diff",
            clearable=False,
            value=0,
            placeholder="Differencing",
            searchable=False,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
        # MA order dropdown
        html.Label("MA Order", htmlFor="model-ma"),
        dcc.Dropdown(
            id="model-ma",
            clearable=False,
            placeholder="MA order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
    ],
)

graphs_and_guide = html.Div(
    className="graphs-and-guide",
    children=[
        dcc.Loading(
            color="#777",
            children=[
                dcc.Graph(id="component-plots"),
            ],
        ),
        html.Div(
            className="line-plot-and-summary",
            children=[
                dcc.Loading(
                    color="#777",
                    children=[
                        dcc.Graph(
                            id="line-plot",
                            config={"toImageButtonOptions": {"format": "svg"}},
                        ),
                    ],
                ),
                dcc.Markdown(summary, className="guide"),
                html.Div(
                    className="footer",
                    children=[
                        html.A(
                            "Back to home",
                            href="/",
                            className="button",
                        ),
                        html.A(
                            "Browse glossary",
                            href="/glossary",
                            className="button",
                        ),
                    ],
                ),
            ],
        ),
    ],
)


@app.callback(
    [
        Output("line-plot", "figure"),
        Output("component-plots", "figure"),
    ],
    [
        Input("model-ar", "value"),
        Input("model-diff", "value"),
        Input("model-ma", "value"),
        Input("url", "pathname"),
        Input("sample-data-store", "data"),
        Input("file-upload-store", "data"),
    ],
)
def model_and_predict(
    ar_order: int,
    diff_order: int,
    ma_order: int,
    input_source: str,
    sample: Optional[dict],
    upload: Optional[dict],
) -> Tuple[go.Figure, go.Figure]:
    """Fit an ARIMA model each time model parameters or input data are
    modified, then plot the results.

    Args:
        ar_order (int): AR order.
        diff_order (int): Differencing order.
        ma_order (int): MA order.
        input_source (str): The data source.
        sample (Optional[dict]): Stored sample data, if any.
        upload (Optional[dict]): Uploaded data, if any.

    Returns:
        Tuple[plotly.graph_objs._figure.Figure, ...]: A line-plot of the
        forecast results, and subplots with seasonal decomposition estimates.
    """
    if input_source == "/upload" and upload is not None:
        filename = upload["filename"]
        data = pd.read_json(upload["data"], orient="index", typ="series")
    elif input_source == "/sample" and sample is not None:
        filename = sample["filename"]
        data = pd.read_json(sample["data"], orient="index", typ="series")
    else:
        filename = "a random sample"
        data = create_arma_sample()

    predictions, forecast = fit_arima_model(
        data, ar_order, diff_order, ma_order
    )

    line_plot = plotting.plot_forecast(
        actual_data=data,
        predictions=predictions,
        forecast=forecast,
        model_info=f"ARIMA({ar_order}, {diff_order}, {ma_order})",
        file_name=filename,
    )

    components = seasonal_decompose(data)
    component_subplots = plotting.plot_ts_components(
        trend=components.trend,
        seasonal=components.seasonal,
        residuals=components.resid,
        file_name=filename,
    )

    return line_plot, component_subplots



import json
from pkgutil import get_data

from dash import html

glossary_data = get_data("ts_app", "assets/glossary.json")
definitions = json.loads(glossary_data.decode())

title_definition_pairs = sum(  # Concatenate
    [
        [
            html.Dt(item["title"], id=item["title"]),
            html.Dd(item["definition"], id=item["definition"]),
        ]
        for item in sorted(definitions, key=lambda x: x["title"])
    ],
    start=[],
)

layout = html.Div(
    [
        html.Div(
            className="glossary-content",
            children=[
                html.H2("Glossary of Time Series Terms"),
                html.Dl(title_definition_pairs),
            ],
        ),
        html.Div(
            className="footer",
            children=[
                html.A("Back to home", href="/", className="button")
            ],
        ),
    ]
)



import warnings
from datetime import date
from typing import Tuple

import numpy as np
import pandas as pd
from statsmodels.api import tsa

# Ignore warnings from statsmodels. `ConvergenceWarning`s and `ValueWarning`s
# are all too frequent when fitting models on arbitrary data.
warnings.filterwarnings("ignore", module="statsmodels")


def create_arma_sample(
    ar_order: int = 1, ma_order: int = 1, size: int = 100
) -> pd.Series:
    """Get a random ARMA sample.

    Args:
        ar_order (int, optional): Desired AR order. Defaults to 1.
        ma_order (int, optional): Desired MA order. Defaults to 1.
        size (int, optional): Desired sample size. Defaults to 100.

    Returns:
        pandas.Series: An ARMA sample.
    """
    ar_coeff = np.linspace(1, -0.9, ar_order + 1)  # arbitrary ar coefficients
    ma_coeff = np.linspace(1, 0.9, ma_order + 1)  # arbitrary ma coefficients
    sample = tsa.ArmaProcess(ar_coeff, ma_coeff).generate_sample(size)
    index = pd.date_range(start=date.today(), periods=size, freq="D")

    return pd.Series(sample, index=index, name="sample")


def fit_arima_model(
    data: pd.Series, ar_order: int = 1, diff: int = 0, ma_order: int = 1
) -> Tuple[pd.Series, pd.Series]:
    """Fit an ARIMA model on the data and get predictions.

    Args:
        data (pandas.Series): The data to model, with a DatetimeIndex.
        ar_order (int, optional): AR order. Defaults to 1.
        diff (int, optional): Differencing order. Defaults to 0.
        ma_order (int, optional): MA order. Defaults to 1.

    Returns:
        Tuple[pandas.Series, pandas.Series]: In-sample predictions covering
        the latter 30% of the data, and a 14-period out-of-sample forecast.
    """
    arima_model = tsa.arima.ARIMA(data, order=(ar_order, diff, ma_order)).fit()
    n = len(data)
    predictions = arima_model.predict(start=int(0.7 * n), end=n)
    forecast = arima_model.predict(start=n, end=n + 14)

    return predictions, forecast



import argparse


def process_cli_args() -> None:
    """Create arguments and collect input from the command line."""
    parser = argparse.ArgumentParser(
        prog="ts_app",
        description=(
            "A simple dashboard application to learn time series basics and"
            " interactively fit ARIMA models."
        ),
    )
    parser.add_argument(
        "-p",
        "--port",
        default=8000,
        type=int,
        help="The TCP port on which to listen (default: %(default)s).",
    )
    parser.add_argument(
        "--host",
        default="0.0.0.0",
        help="A host-name or IP address (default: %(default)r).",
    )
    parser.add_argument(
        "--no-browser",
        action="store_true",
        help="Avoid openning a browser tab or window.",
    )
    return parser.parse_args()



import logging
import webbrowser

from waitress import serve

from ts_app.cli import process_cli_args
from ts_app.dashboard import server

__version__ = "0.8.0"

logging.basicConfig(level="INFO")

# Suppress benign queue warnings
logging.getLogger("waitress.queue").setLevel("ERROR")


def run_app(
    host: str = "0.0.0.0", port: int = 8000, launch_browser: bool = True
) -> None:
    """Start the app server, and launch a web browser to it.

    Args:
        host (str, optional): A host-name/IP address. Defaults to "0.0.0.0".
        port (int, optional): TCP port to listen at. Defaults to 8000.
        launch_browser (bool, optional): Whether to launch a web browser to
            view the app. Defaults to True.
    """
    if launch_browser is True:
        webbrowser.open(f"{host}:{port}")

    serve(server, host=host, port=port)


def _run_in_cli() -> None:
    """Start the app server using input collected from the command line
    interface.
    """
    args = process_cli_args()

    if not args.no_browser:
        webbrowser.open(f"{args.host}:{args.port}")

    serve(server, host=args.host.split("://")[-1], port=args.port)



from pkgutil import get_data

from dash import dcc, html

page_content = get_data("ts_app", "assets/home-page.md").decode("utf-8")

layout = html.Div(
    className="home-page-content",
    children=[
        # Introduction
        dcc.Markdown(page_content),
        html.Div(
            className="footer",
            children=[
                html.A(
                    "Upload a file", href="/upload", className="button"
                ),
                html.A(
                    "Process a sample",
                    href="/sample",
                    className="button",
                ),
            ],
        ),
    ],
)



from base64 import b64decode
from io import BytesIO, StringIO
from typing import Optional, Tuple

import pandas as pd
from dash import dcc, html
from dash.dependencies import Input, Output
from ts_app.dash_app import app
from ts_app.file_upload import process_upload

input_layout = html.Div(
    [
        dcc.Upload(
            id="file-upload",
            accept=".csv,.xls,.xlsx",
            className="file-upload",
            children=[
                "Click or Drag and Drop",
                html.P("Expected file properties:"),
                html.Ul(
                    children=[
                        html.Li("At most 7MiB"),
                        html.Li("Dates in first column"),
                        html.Li("Numeric data in right-most column"),
                        html.Li("At least 32 rows"),
                    ]
                ),
            ],
            min_size=32,
            max_size=1024**2 * 7,  # 7MiB
        ),
        html.P(id="file-info"),
    ]
)


@app.callback(
    [
        Output("file-info", "children"),
        Output("file-info", "style"),
        Output("file-upload-store", "data"),
    ],
    [Input("file-upload", "contents"), Input("file-upload", "filename")],
)
def upload_file(
    contents: str, filename: str
) -> Tuple[str, dict, Optional[dict]]:
    """Extract and validate data from uploaded files.

    Args:
        contents (str): Base64-encoded string with the file's contents.
        filename (str): The name of the uploaded file.

    Returns:
        Tuple[str, dict, Optional[dict]]: file-info message, file-info style
        and the data-dict to store.
    """
    if contents is None:
        return (
            "",  # No file information
            {},  # No special style
            None,  # No data to store
        )
    else:
        content_string = contents.split(",")[1]
        file_content = b64decode(content_string)

    try:
        if filename.endswith(".csv"):
            df = pd.read_csv(
                StringIO(file_content.decode("utf-8")), index_col=0
            )
        elif filename.endswith(".xls") or filename.endswith(".xlsx"):
            df = pd.read_excel(BytesIO(file_content), index_col=0)
    except Exception as error:
        print(error)
        return (
            html.Div(["There was an error processing the file."]),
            {"color": "orangered"},
            None,  # No data to store
        )

    if (validation_error := process_upload(data=df)) is not None:
        return (
            validation_error,
            {"color": "orangered"},
            None,  # No data to store
        )
    else:
        # If file-upload and data-extraction succeed
        return (
            f"Analysing {filename}",
            {"color": "#31bf2c"},
            {"filename": filename, "data": df.iloc[:, -1].to_json()},
        )
